use palette::Hsl;
use serde::Serialize;
use wasm_bindgen::prelude::wasm_bindgen;

use super::{ColorDifference, Differ};

#[derive(Debug, Clone, Copy, Serialize)]
#[wasm_bindgen]
pub struct HSLDifference {
    pub hue: Differ,
    pub saturation: Differ,
    pub lightness: Differ,
}

#[wasm_bindgen]
impl HSLDifference {
    #[wasm_bindgen(constructor)]
    pub fn new(hue: Differ, saturation: Differ, lightness: Differ) -> Self {
        Self {
            hue,
            saturation,
            lightness,
        }
    }
}

impl ColorDifference for Hsl {
    type Difference = HSLDifference;

    fn difference(&self, other: &Self) -> Self::Difference {
        let hue = other.hue.into_positive_degrees() - self.hue.into_positive_degrees();
        let saturation = other.saturation - self.saturation;
        let lightness = other.lightness - self.lightness;

        HSLDifference {
            hue: Differ {
                delta: hue,
                percent: hue / self.hue.into_positive_degrees(),
            },
            saturation: Differ {
                delta: saturation,
                percent: saturation / self.saturation,
            },
            lightness: Differ {
                delta: lightness,
                percent: lightness / self.lightness,
            },
        }
    }

    fn difference_relative(&self, other: &Self) -> Self::Difference {
        let hue = if self.hue.into_positive_degrees() == 0.0 {
            0.0
        } else {
            (other.hue.into_positive_degrees() - self.hue.into_positive_degrees())
                / (1.0 - self.hue.into_positive_degrees())
        };
        let saturation = if self.saturation == 0.0 {
            0.0
        } else {
            (other.saturation - self.saturation) / (1.0 - self.saturation)
        };
        let lightness = if self.lightness == 0.0 {
            0.0
        } else {
            (other.lightness - self.lightness) / (1.0 - self.lightness)
        };

        HSLDifference {
            hue: Differ {
                delta: other.hue.into_positive_degrees() - self.hue.into_positive_degrees(),
                percent: hue,
            },
            saturation: Differ {
                delta: other.saturation - self.saturation,
                percent: saturation,
            },
            lightness: Differ {
                delta: other.lightness - self.lightness,
                percent: lightness,
            },
        }
    }
}
