use linked_hash_map::LinkedHashMap;
use serde::Serialize;

use crate::{convert::map_hsl_to_srgb_hex, errors};

use super::swatch::M2Swatch;

#[derive(Debug, Clone, Serialize)]
pub struct M2ColorSet {
    pub root: String,
    pub variant: String,
    pub on: String,
}

impl M2ColorSet {
    pub fn from_swatch(
        swatch: &M2Swatch,
        neutral: &M2Swatch,
        dark: bool,
        required_wacg_ratio: Option<f32>,
    ) -> Result<Self, errors::ColorError> {
        let root_color = if dark {
            map_hsl_to_srgb_hex(&swatch.desaturated_key_tone())
        } else {
            map_hsl_to_srgb_hex(&swatch.key_tone())
        };
        if dark {
            Ok(Self {
                variant: map_hsl_to_srgb_hex(&swatch.desaturated_tone(&swatch.key_index - 2)),
                on: match required_wacg_ratio {
                    Some(ratio) => map_hsl_to_srgb_hex(
                        &neutral.desaturated_wacg_min_above(&root_color, ratio)?,
                    ),
                    None => map_hsl_to_srgb_hex(&neutral.desaturated_wacg_max_tone(&root_color)?),
                },
                root: root_color,
            })
        } else {
            Ok(Self {
                variant: map_hsl_to_srgb_hex(&swatch.tone(&swatch.key_index + 2)),
                on: match required_wacg_ratio {
                    Some(ratio) => {
                        map_hsl_to_srgb_hex(&neutral.wacg_min_above(&root_color, ratio)?)
                    }
                    None => map_hsl_to_srgb_hex(&neutral.wacg_max_tone(&root_color)?),
                },
                root: root_color,
            })
        }
    }

    pub fn surface(neutral: &M2Swatch, dark: bool) -> Result<Self, errors::ColorError> {
        let root_color_index = if dark {
            super::swatch::SwatchIndex::SI900
        } else {
            super::swatch::SwatchIndex::SI50
        };
        if dark {
            Ok(Self {
                root: map_hsl_to_srgb_hex(&neutral.desaturated_tone(root_color_index)),
                variant: map_hsl_to_srgb_hex(&neutral.desaturated_tone(root_color_index + 2)),
                on: map_hsl_to_srgb_hex(&neutral.tone(super::swatch::SwatchIndex::SI50)),
            })
        } else {
            Ok(Self {
                root: map_hsl_to_srgb_hex(&neutral.tone(root_color_index)),
                variant: map_hsl_to_srgb_hex(&neutral.tone(root_color_index - 2)),
                on: map_hsl_to_srgb_hex(&neutral.tone(super::swatch::SwatchIndex::SI900)),
            })
        }
    }

    pub fn to_css_variable(&self, prefix: &str, name: &str) -> Vec<String> {
        let mut variable_lines = Vec::new();

        variable_lines.push(format!("--color-{}-{}: #{};", prefix, name, self.root));
        variable_lines.push(format!(
            "--color-{}-{}-variant: #{};",
            prefix, name, self.variant
        ));
        variable_lines.push(format!("--color-{}-on-{}: #{};", prefix, name, self.on));

        variable_lines
    }

    pub fn to_css_auto_scheme_collection(&self, name: &str) -> LinkedHashMap<String, String> {
        let mut variables = LinkedHashMap::new();

        variables.insert(format!("{}", name), self.root.clone());
        variables.insert(format!("{}-variant", name), self.variant.clone());
        variables.insert(format!("on-{}", name), self.on.clone());

        variables
    }

    pub fn to_scss_variable(&self, prefix: &str, name: &str) -> Vec<String> {
        let mut variable_lines = Vec::new();

        variable_lines.push(format!("$color-{}-{}: #{};", prefix, name, self.root));
        variable_lines.push(format!(
            "$color-{}-{}-variant: #{};",
            prefix, name, self.variant
        ));
        variable_lines.push(format!("$color-{}-on-{}: #{};", prefix, name, self.on));

        variable_lines
    }

    pub fn to_javascript_object(&self, prefix: &str, name: &str) -> Vec<String> {
        let mut variable_lines = Vec::new();
        let prefix = prefix.to_ascii_lowercase();
        let name = name
            .chars()
            .next()
            .unwrap()
            .to_ascii_uppercase()
            .to_string()
            + &name[1..];

        variable_lines.push(format!("{}{}: '#{}',", prefix, name, self.root));
        variable_lines.push(format!("{}{}Variant: '#{}',", prefix, name, self.variant));
        variable_lines.push(format!("{}On{}: '#{}',", prefix, name, self.on));

        variable_lines
    }
}
