use std::collections::HashMap;

use serde::Serialize;

use crate::convert::map_lch_to_srgb_hex;

use super::TonalPalette;

static SWATCH_TONES: [u8; 18] = [
    0, 5, 10, 15, 20, 25, 30, 35, 40, 50, 60, 70, 80, 90, 95, 98, 99, 100,
];

#[derive(Debug, Clone, Serialize)]
pub struct M3PaletteSwatch(HashMap<u8, String>);

impl M3PaletteSwatch {
    pub fn new(palette: &TonalPalette) -> Self {
        let mut swatch = HashMap::new();
        for &tone in SWATCH_TONES.iter() {
            let color = palette.tone(tone as f32);
            swatch.insert(tone, map_lch_to_srgb_hex(&color));
        }
        Self(swatch)
    }

    pub fn to_css_variables(&self, name: &str) -> Vec<String> {
        let mut variable_lines = Vec::new();
        let name = name.replace('_', "-").to_lowercase();

        for &tone in SWATCH_TONES.iter() {
            let color = self.0.get(&tone).unwrap();
            variable_lines.push(format!("--color-swatch-{}-{}: #{};", name, tone, color));
        }

        variable_lines
    }

    pub fn to_scss_variables(&self, name: &str) -> Vec<String> {
        let mut variable_lines = Vec::new();
        let name = name.replace('_', "-").to_lowercase();

        for &tone in SWATCH_TONES.iter() {
            let color = self.0.get(&tone).unwrap();
            variable_lines.push(format!("$color-swatch-{}-{}: #{};", name, tone, color));
        }

        variable_lines
    }

    pub fn to_javascript_object_fields(&self, name: &str) -> Vec<String> {
        let mut js_object = Vec::new();
        let name = name
            .split('_')
            .enumerate()
            .map(|(i, part)| {
                if i == 0 {
                    part.to_string()
                } else {
                    let mut c = part.chars();
                    c.next().unwrap().to_uppercase().collect::<String>() + c.as_str()
                }
            })
            .collect::<String>();

        js_object.push(format!("{}: {{", name));
        for &tone in SWATCH_TONES.iter() {
            let color = self.0.get(&tone).unwrap();
            js_object.push(format!("  {}: '#{}',", tone, color));
        }
        js_object.push("},".to_string());

        js_object
    }
}
