use enum_iterator::Sequence;
use palette::{
    color_theory::{Analogous, Complementary},
    Lch,
};
use serde_repr::{Deserialize_repr, Serialize_repr};
use strum::Display;
use wasm_bindgen::prelude::wasm_bindgen;

use crate::{
    schemes::material_design_3::TonalPalette,
    theory::{harmonize_hue, sanitize_hue_degrees},
};

use super::dynamic_color::CustomPaletteGenerator;

#[derive(Debug, Clone, Copy, PartialEq, Display, Sequence, Serialize_repr, Deserialize_repr)]
#[wasm_bindgen]
#[repr(u8)]
pub enum Variant {
    Monochrome,
    Neutral,
    TonalSpot,
    Vibrant,
    Expressive,
    Fidelity,
    Content,
    Rainbow,
    FruitSalad,
}

impl Variant {
    pub fn label(&self) -> String {
        match self {
            Variant::Monochrome => "Monochrome".to_string(),
            Variant::Neutral => "Neutral".to_string(),
            Variant::TonalSpot => "Tonal Spot".to_string(),
            Variant::Vibrant => "Vibrant".to_string(),
            Variant::Expressive => "Expressive".to_string(),
            Variant::Fidelity => "Fidelity".to_string(),
            Variant::Content => "Content".to_string(),
            Variant::Rainbow => "Rainbow".to_string(),
            Variant::FruitSalad => "Fruit Salad".to_string(),
        }
    }

    pub fn from_u8(value: u8) -> Variant {
        match value {
            0 => Variant::Monochrome,
            1 => Variant::Neutral,
            2 => Variant::TonalSpot,
            3 => Variant::Vibrant,
            4 => Variant::Expressive,
            5 => Variant::Fidelity,
            6 => Variant::Content,
            7 => Variant::Rainbow,
            8 => Variant::FruitSalad,
            _ => Variant::Expressive,
        }
    }

    pub fn hues(&self) -> Vec<f32> {
        match self {
            Variant::Vibrant => vec![0.0, 41.0, 61.0, 101.0, 131.0, 181.0, 251.0, 301.0, 360.0],
            Variant::Expressive => vec![0.0, 21.0, 51.0, 121.0, 151.0, 191.0, 271.0, 321.0, 360.0],
            _ => vec![],
        }
    }

    pub fn secondary_rotation(&self) -> Vec<f32> {
        match self {
            Variant::Vibrant => vec![18.0, 15.0, 10.0, 12.0, 15.0, 18.0, 15.0, 12.0, 12.0],
            Variant::Expressive => vec![45.0, 95.0, 45.0, 20.0, 45.0, 90.0, 45.0, 45.0, 45.0],
            _ => vec![],
        }
    }

    pub fn tertiary_rotation(&self) -> Vec<f32> {
        match self {
            Variant::Vibrant => vec![35.0, 30.0, 20.0, 25.0, 30.0, 35.0, 30.0, 25.0, 25.0],
            Variant::Expressive => vec![120.0, 120.0, 20.0, 45.0, 20.0, 15.0, 20.0, 120.0, 120.0],
            _ => vec![],
        }
    }

    pub fn build_palette(
        &self,
        source_color: Lch,
        harmonize_customs: bool,
    ) -> (
        TonalPalette,
        TonalPalette,
        TonalPalette,
        TonalPalette,
        TonalPalette,
        CustomPaletteGenerator,
    ) {
        match self {
            Variant::Monochrome => (
                TonalPalette::from_hue_and_chroma(source_color.hue.into_positive_degrees(), 0.0),
                TonalPalette::from_hue_and_chroma(source_color.hue.into_positive_degrees(), 0.0),
                TonalPalette::from_hue_and_chroma(source_color.hue.into_positive_degrees(), 0.0),
                TonalPalette::from_hue_and_chroma(source_color.hue.into_positive_degrees(), 0.0),
                TonalPalette::from_hue_and_chroma(source_color.hue.into_positive_degrees(), 0.0),
                if harmonize_customs {
                    let source_hue = source_color.hue.into_positive_degrees();
                    Box::new(move |c: &Lch| {
                        TonalPalette::from_hue_and_chroma(
                            harmonize_hue(c.hue.into_positive_degrees(), source_hue),
                            0.0,
                        )
                    })
                } else {
                    Box::new(|c: &Lch| {
                        TonalPalette::from_hue_and_chroma(c.hue.into_positive_degrees(), 0.0)
                    })
                },
            ),
            Variant::Neutral => (
                TonalPalette::from_hue_and_chroma(source_color.hue.into_positive_degrees(), 12.0),
                TonalPalette::from_hue_and_chroma(source_color.hue.into_positive_degrees(), 8.0),
                TonalPalette::from_hue_and_chroma(source_color.hue.into_positive_degrees(), 16.0),
                TonalPalette::from_hue_and_chroma(source_color.hue.into_positive_degrees(), 2.0),
                TonalPalette::from_hue_and_chroma(source_color.hue.into_positive_degrees(), 2.0),
                if harmonize_customs {
                    let source_hue = source_color.hue.into_positive_degrees();
                    Box::new(move |c: &Lch| {
                        TonalPalette::from_hue_and_chroma(
                            harmonize_hue(c.hue.into_positive_degrees(), source_hue),
                            12.0,
                        )
                    })
                } else {
                    Box::new(|c: &Lch| {
                        TonalPalette::from_hue_and_chroma(c.hue.into_positive_degrees(), 12.0)
                    })
                },
            ),
            Variant::TonalSpot => (
                TonalPalette::from_hue_and_chroma(source_color.hue.into_positive_degrees(), 36.0),
                TonalPalette::from_hue_and_chroma(source_color.hue.into_positive_degrees(), 16.0),
                TonalPalette::from_hue_and_chroma(
                    sanitize_hue_degrees(source_color.hue.into_positive_degrees() + 60.0),
                    24.0,
                ),
                TonalPalette::from_hue_and_chroma(source_color.hue.into_positive_degrees(), 6.0),
                TonalPalette::from_hue_and_chroma(source_color.hue.into_positive_degrees(), 8.0),
                if harmonize_customs {
                    let source_hue = source_color.hue.into_positive_degrees();
                    Box::new(move |c: &Lch| {
                        TonalPalette::from_hue_and_chroma(
                            harmonize_hue(c.hue.into_positive_degrees(), source_hue),
                            36.0,
                        )
                    })
                } else {
                    Box::new(|c: &Lch| {
                        TonalPalette::from_hue_and_chroma(c.hue.into_positive_degrees(), 36.0)
                    })
                },
            ),
            Variant::Vibrant => (
                TonalPalette::from_hue_and_chroma(source_color.hue.into_positive_degrees(), 200.0),
                TonalPalette::from_hue_and_chroma(
                    rotate_hue(&source_color, &self.hues(), &self.secondary_rotation()),
                    24.0,
                ),
                TonalPalette::from_hue_and_chroma(
                    rotate_hue(&source_color, &self.hues(), &self.tertiary_rotation()),
                    32.0,
                ),
                TonalPalette::from_hue_and_chroma(source_color.hue.into_positive_degrees(), 10.0),
                TonalPalette::from_hue_and_chroma(source_color.hue.into_positive_degrees(), 12.0),
                if harmonize_customs {
                    let source_hue = source_color.hue.into_positive_degrees();
                    Box::new(move |c: &Lch| {
                        TonalPalette::from_hue_and_chroma(
                            harmonize_hue(c.hue.into_positive_degrees(), source_hue),
                            200.0,
                        )
                    })
                } else {
                    Box::new(|c: &Lch| {
                        TonalPalette::from_hue_and_chroma(c.hue.into_positive_degrees(), 200.0)
                    })
                },
            ),
            Variant::Expressive => (
                TonalPalette::from_hue_and_chroma(
                    sanitize_hue_degrees(source_color.hue.into_positive_degrees() + 240.0),
                    40.0,
                ),
                TonalPalette::from_hue_and_chroma(
                    rotate_hue(&source_color, &self.hues(), &self.secondary_rotation()),
                    24.0,
                ),
                TonalPalette::from_hue_and_chroma(
                    rotate_hue(&source_color, &self.hues(), &self.tertiary_rotation()),
                    32.0,
                ),
                TonalPalette::from_hue_and_chroma(
                    source_color.hue.into_positive_degrees() + 15.0,
                    8.0,
                ),
                TonalPalette::from_hue_and_chroma(
                    source_color.hue.into_positive_degrees() + 15.0,
                    12.0,
                ),
                if harmonize_customs {
                    let source_hue =
                        sanitize_hue_degrees(source_color.hue.into_positive_degrees() + 240.0);
                    Box::new(move |c: &Lch| {
                        TonalPalette::from_hue_and_chroma(
                            harmonize_hue(c.hue.into_positive_degrees(), source_hue),
                            40.0,
                        )
                    })
                } else {
                    Box::new(|c: &Lch| {
                        TonalPalette::from_hue_and_chroma(c.hue.into_positive_degrees(), 40.0)
                    })
                },
            ),
            Variant::Fidelity => (
                TonalPalette::from_hue_and_chroma(
                    source_color.hue.into_positive_degrees(),
                    source_color.chroma,
                ),
                TonalPalette::from_hue_and_chroma(
                    source_color.hue.into_positive_degrees(),
                    (source_color.chroma - 32.0).max(source_color.chroma * 0.5),
                ),
                TonalPalette {
                    key_color: fix_disliked(&source_color.complementary()),
                },
                TonalPalette::from_hue_and_chroma(
                    source_color.hue.into_positive_degrees(),
                    source_color.chroma / 8.0,
                ),
                TonalPalette::from_hue_and_chroma(
                    source_color.hue.into_positive_degrees(),
                    source_color.chroma / 8.0 + 4.0,
                ),
                if harmonize_customs {
                    let source_hue = source_color.hue.into_positive_degrees();
                    let source_chroma = source_color.chroma;
                    Box::new(move |c: &Lch| {
                        TonalPalette::from_hue_and_chroma(
                            harmonize_hue(c.hue.into_positive_degrees(), source_hue),
                            source_chroma,
                        )
                    })
                } else {
                    let source_chroma = source_color.chroma;
                    Box::new(move |c: &Lch| {
                        TonalPalette::from_hue_and_chroma(
                            c.hue.into_positive_degrees(),
                            source_chroma,
                        )
                    })
                },
            ),
            Variant::Content => (
                TonalPalette::from_hue_and_chroma(
                    source_color.hue.into_positive_degrees(),
                    source_color.chroma,
                ),
                TonalPalette::from_hue_and_chroma(
                    source_color.hue.into_positive_degrees(),
                    (source_color.chroma - 32.0).max(source_color.chroma * 0.5),
                ),
                TonalPalette {
                    key_color: fix_disliked(&source_color.analogous().1),
                },
                TonalPalette::from_hue_and_chroma(
                    source_color.hue.into_positive_degrees(),
                    source_color.chroma / 8.0,
                ),
                TonalPalette::from_hue_and_chroma(
                    source_color.hue.into_positive_degrees(),
                    source_color.chroma / 8.0 + 4.0,
                ),
                if harmonize_customs {
                    let source_hue = source_color.hue.into_positive_degrees();
                    Box::new(move |c: &Lch| {
                        TonalPalette::from_hue_and_chroma(
                            harmonize_hue(c.hue.into_positive_degrees(), source_hue),
                            c.chroma,
                        )
                    })
                } else {
                    Box::new(|c: &Lch| {
                        TonalPalette::from_hue_and_chroma(c.hue.into_positive_degrees(), c.chroma)
                    })
                },
            ),
            Variant::Rainbow => (
                TonalPalette::from_hue_and_chroma(source_color.hue.into_positive_degrees(), 48.0),
                TonalPalette::from_hue_and_chroma(source_color.hue.into_positive_degrees(), 16.0),
                TonalPalette::from_hue_and_chroma(
                    sanitize_hue_degrees(source_color.hue.into_positive_degrees() + 60.0),
                    24.0,
                ),
                TonalPalette::from_hue_and_chroma(source_color.hue.into_positive_degrees(), 0.0),
                TonalPalette::from_hue_and_chroma(source_color.hue.into_positive_degrees(), 0.0),
                if harmonize_customs {
                    let source_hue = source_color.hue.into_positive_degrees();
                    Box::new(move |c: &Lch| {
                        TonalPalette::from_hue_and_chroma(
                            harmonize_hue(c.hue.into_positive_degrees(), source_hue),
                            48.0,
                        )
                    })
                } else {
                    Box::new(|c: &Lch| {
                        TonalPalette::from_hue_and_chroma(c.hue.into_positive_degrees(), 48.0)
                    })
                },
            ),
            Variant::FruitSalad => (
                TonalPalette::from_hue_and_chroma(
                    sanitize_hue_degrees(source_color.hue.into_positive_degrees() - 50.0),
                    48.0,
                ),
                TonalPalette::from_hue_and_chroma(
                    sanitize_hue_degrees(source_color.hue.into_positive_degrees() - 50.0),
                    36.0,
                ),
                TonalPalette::from_hue_and_chroma(source_color.hue.into_positive_degrees(), 36.0),
                TonalPalette::from_hue_and_chroma(source_color.hue.into_positive_degrees(), 10.0),
                TonalPalette::from_hue_and_chroma(source_color.hue.into_positive_degrees(), 16.0),
                if harmonize_customs {
                    let source_hue =
                        sanitize_hue_degrees(source_color.hue.into_positive_degrees() - 50.0);
                    Box::new(move |c: &Lch| {
                        TonalPalette::from_hue_and_chroma(
                            harmonize_hue(c.hue.into_positive_degrees(), source_hue),
                            48.0,
                        )
                    })
                } else {
                    Box::new(|c: &Lch| {
                        TonalPalette::from_hue_and_chroma(c.hue.into_positive_degrees(), 48.0)
                    })
                },
            ),
        }
    }
}

fn rotate_hue(source: &Lch, hues: &Vec<f32>, rotations: &Vec<f32>) -> f32 {
    let source_hue = source.hue.into_positive_degrees();
    if rotations.len() == 1 {
        return sanitize_hue_degrees(source_hue + rotations[0]);
    }
    let hues_size = hues.len();
    for i in 0..=hues_size - 2 {
        let hue = hues[i];
        let next_hue = hues[i + 1];
        if hue < source_hue && source_hue < next_hue {
            return sanitize_hue_degrees(source_hue + rotations[i]);
        }
    }

    source_hue
}

pub fn fix_disliked(color: &Lch) -> Lch {
    let hue = color.hue.into_positive_degrees().round() >= 90.0
        && color.hue.into_positive_degrees().round() <= 111.0;
    let chroma = color.chroma.round() > 16.0;
    let lightness = color.l.round() < 65.0;

    if hue && chroma && lightness {
        Lch::new(70.0, color.chroma, color.hue)
    } else {
        color.clone()
    }
}
