use std::{collections::HashMap, str::FromStr};

use dynamic_scheme::DynamicScheme;
use material_colors::*;
use palette::{IntoColor, Lch, Srgb};
use wasm_bindgen::{prelude::wasm_bindgen, JsValue};

use crate::{cond, convert::map_lch_to_srgb_hex, errors};

use super::material_design_3::{M3BaselineColors, M3ColorSet, M3PaletteSwatch, M3SurfaceSet};
pub use constants::Variant;

mod constants;
mod contrast;
mod contrast_curve;
mod dynamic_color;
mod dynamic_scheme;
mod material_colors;

#[wasm_bindgen]
pub fn material_design_3_dynamic_variant() -> Result<JsValue, String> {
    let variants = enum_iterator::all::<constants::Variant>()
        .map(|variant| {
            serde_json::json!({
                "label": variant.label(),
                "value": variant as u8,
            })
        })
        .collect::<Vec<_>>();

    Ok(serde_wasm_bindgen::to_value(&variants).map_err(|e| e.to_string())?)
}

fn build_primary_color_set(scheme: &DynamicScheme) -> M3ColorSet {
    M3ColorSet {
        root: map_lch_to_srgb_hex(&PRIMARY.get_lch(scheme)),
        on_root: map_lch_to_srgb_hex(&ON_PRIMARY.get_lch(scheme)),
        container: map_lch_to_srgb_hex(&PRIMARY_CONTAINER.get_lch(scheme)),
        on_container: map_lch_to_srgb_hex(&ON_PRIMARY_CONTAINER.get_lch(scheme)),
        fixed: map_lch_to_srgb_hex(&PRIMARY_FIXED.get_lch(scheme)),
        on_fixed: map_lch_to_srgb_hex(&ON_PRIMARY_FIXED.get_lch(scheme)),
        fixed_variant: map_lch_to_srgb_hex(&ON_PRIMARY_FIXED_VARIANT.get_lch(scheme)),
        fixed_dim: map_lch_to_srgb_hex(&PRIMARY_FIXED_DIM.get_lch(scheme)),
        inverse: map_lch_to_srgb_hex(&INVERSE_PRIMARY.get_lch(scheme)),
    }
}

fn build_secondary_color_set(scheme: &DynamicScheme) -> M3ColorSet {
    M3ColorSet {
        root: map_lch_to_srgb_hex(&SECONDARY.get_lch(scheme)),
        on_root: map_lch_to_srgb_hex(&ON_SECONDARY.get_lch(scheme)),
        container: map_lch_to_srgb_hex(&SECONDARY_CONTAINER.get_lch(scheme)),
        on_container: map_lch_to_srgb_hex(&ON_SECONDARY_CONTAINER.get_lch(scheme)),
        fixed: map_lch_to_srgb_hex(&SECONDARY_FIXED.get_lch(scheme)),
        on_fixed: map_lch_to_srgb_hex(&ON_SECONDARY_FIXED.get_lch(scheme)),
        fixed_variant: map_lch_to_srgb_hex(&ON_SECONDARY_FIXED_VARIANT.get_lch(scheme)),
        fixed_dim: map_lch_to_srgb_hex(&SECONDARY_FIXED_DIM.get_lch(scheme)),
        ..cond!(
            scheme.is_dark,
            M3ColorSet::new_dark_set(&scheme.secondary_palette),
            M3ColorSet::new_light_set(&scheme.secondary_palette)
        )
    }
}

fn build_tertiary_color_set(scheme: &DynamicScheme) -> M3ColorSet {
    M3ColorSet {
        root: map_lch_to_srgb_hex(&TERTIARY.get_lch(scheme)),
        on_root: map_lch_to_srgb_hex(&ON_TERTIARY.get_lch(scheme)),
        container: map_lch_to_srgb_hex(&TERTIARY_CONTAINER.get_lch(scheme)),
        on_container: map_lch_to_srgb_hex(&ON_TERTIARY_CONTAINER.get_lch(scheme)),
        fixed: map_lch_to_srgb_hex(&TERTIARY_FIXED.get_lch(scheme)),
        on_fixed: map_lch_to_srgb_hex(&ON_TERTIARY_FIXED.get_lch(scheme)),
        fixed_variant: map_lch_to_srgb_hex(&ON_TERTIARY_FIXED_VARIANT.get_lch(scheme)),
        fixed_dim: map_lch_to_srgb_hex(&TERTIARY_FIXED_DIM.get_lch(scheme)),
        ..cond!(
            scheme.is_dark,
            M3ColorSet::new_dark_set(&scheme.tertiary_palette),
            M3ColorSet::new_light_set(&scheme.tertiary_palette)
        )
    }
}

fn build_surface_color_set(scheme: &DynamicScheme) -> M3SurfaceSet {
    M3SurfaceSet {
        root: map_lch_to_srgb_hex(&SURFACE.get_lch(scheme)),
        dim: map_lch_to_srgb_hex(&SURFACE_DIM.get_lch(scheme)),
        bright: map_lch_to_srgb_hex(&SURFACE_BRIGHT.get_lch(scheme)),
        variant: map_lch_to_srgb_hex(&SURFACE_VARIANT.get_lch(scheme)),
        container: map_lch_to_srgb_hex(&SURFACE_CONTAINER.get_lch(scheme)),
        container_lowest: map_lch_to_srgb_hex(&SURFACE_CONTAINER_LOWEST.get_lch(scheme)),
        container_low: map_lch_to_srgb_hex(&SURFACE_CONTAINER_LOW.get_lch(scheme)),
        container_high: map_lch_to_srgb_hex(&SURFACE_CONTAINER_HIGH.get_lch(scheme)),
        container_highest: map_lch_to_srgb_hex(&SURFACE_CONTAINER_HIGHEST.get_lch(scheme)),
        on_root: map_lch_to_srgb_hex(&ON_SURFACE.get_lch(scheme)),
        on_root_variant: map_lch_to_srgb_hex(&ON_SURFACE_VARIANT.get_lch(scheme)),
        inverse: map_lch_to_srgb_hex(&INVERSE_SURFACE.get_lch(scheme)),
        on_inverse: map_lch_to_srgb_hex(&INVERSE_ON_SURFACE.get_lch(scheme)),
    }
}

fn build_error_color_set(scheme: &DynamicScheme) -> M3ColorSet {
    M3ColorSet {
        root: map_lch_to_srgb_hex(&ERROR.get_lch(scheme)),
        on_root: map_lch_to_srgb_hex(&ON_ERROR.get_lch(scheme)),
        container: map_lch_to_srgb_hex(&ERROR_CONTAINER.get_lch(scheme)),
        on_container: map_lch_to_srgb_hex(&ON_ERROR_CONTAINER.get_lch(scheme)),
        ..cond!(
            scheme.is_dark,
            M3ColorSet::new_dark_set(&scheme.error_palette),
            M3ColorSet::new_light_set(&scheme.error_palette)
        )
    }
}

fn build_custom_color_set(scheme: &DynamicScheme, name: String) -> M3ColorSet {
    M3ColorSet {
        root: map_lch_to_srgb_hex(&custom(name.clone()).get_lch(scheme)),
        on_root: map_lch_to_srgb_hex(&on_custom(name.clone()).get_lch(scheme)),
        container: map_lch_to_srgb_hex(&custom_container(name.clone()).get_lch(scheme)),
        on_container: map_lch_to_srgb_hex(&on_custom_container(name.clone()).get_lch(scheme)),
        fixed: map_lch_to_srgb_hex(&custom_fixed(name.clone()).get_lch(scheme)),
        on_fixed: map_lch_to_srgb_hex(&on_custom_fixed(name.clone()).get_lch(scheme)),
        fixed_variant: map_lch_to_srgb_hex(&on_custom_fixed_variant(name.clone()).get_lch(scheme)),
        fixed_dim: map_lch_to_srgb_hex(&custom_fixed_dim(name.clone()).get_lch(scheme)),
        inverse: map_lch_to_srgb_hex(&inverse_custom(name.clone()).get_lch(scheme)),
    }
}

pub fn build_baseline(scheme: &DynamicScheme) -> M3BaselineColors {
    M3BaselineColors::full_custom(
        build_primary_color_set(scheme),
        build_secondary_color_set(scheme),
        build_tertiary_color_set(scheme),
        build_error_color_set(scheme),
        build_surface_color_set(scheme),
        map_lch_to_srgb_hex(&OUTLINE.get_lch(scheme)),
        map_lch_to_srgb_hex(&OUTLINE_VARIANT.get_lch(scheme)),
        map_lch_to_srgb_hex(&SCRIM.get_lch(scheme)),
        map_lch_to_srgb_hex(&SHADOW.get_lch(scheme)),
        scheme
            .custom_palettes
            .keys()
            .map(|name| (name.clone(), build_custom_color_set(scheme, name.clone())))
            .collect(),
        scheme.is_dark,
    )
}

pub fn build_swatches(scheme: &DynamicScheme) -> HashMap<String, M3PaletteSwatch> {
    let mut swatches = HashMap::new();
    swatches.insert(
        "primary".to_string(),
        M3PaletteSwatch::new(&scheme.primary_palette),
    );
    swatches.insert(
        "secondary".to_string(),
        M3PaletteSwatch::new(&scheme.secondary_palette),
    );
    swatches.insert(
        "tertiary".to_string(),
        M3PaletteSwatch::new(&scheme.tertiary_palette),
    );
    swatches.insert(
        "error".to_string(),
        M3PaletteSwatch::new(&scheme.error_palette),
    );
    swatches.insert(
        "neutral".to_string(),
        M3PaletteSwatch::new(&scheme.neutral_palette),
    );
    swatches.insert(
        "neutral_variant".to_string(),
        M3PaletteSwatch::new(&scheme.neutral_variant_palette),
    );

    for (name, palette) in &scheme.custom_palettes {
        swatches.insert(name.clone(), M3PaletteSwatch::new(palette));
    }

    swatches
}

pub fn build_dynamic_scheme(
    source_color: &str,
    error_color: Option<String>,
    custom_colors: HashMap<String, String>,
    variant: constants::Variant,
    contrast_level: f32,
    is_dark: bool,
    harmonize_customs: bool,
) -> Result<DynamicScheme, errors::ColorError> {
    let source_color = Srgb::from_str(source_color)
        .map_err(|_| errors::ColorError::UnrecogniazedRGB(source_color.to_string()))?
        .into_format::<f32>()
        .into_color();
    let error_color = error_color
        .map(|color| Srgb::from_str(&color))
        .transpose()
        .map_err(|_| errors::ColorError::UnrecogniazedRGB("error color".to_string()))?
        .map(|color| color.into_format::<f32>().into_color());
    let custom_colors = custom_colors
        .into_iter()
        .map(|(name, color)| {
            let color = Srgb::from_str(&color)
                .map_err(|_| errors::ColorError::UnrecogniazedRGB(color.clone()))?
                .into_format::<f32>()
                .into_color();
            Ok((name, color))
        })
        .collect::<Result<HashMap<String, Lch>, errors::ColorError>>()?;
    Ok(DynamicScheme::new(
        source_color,
        error_color,
        custom_colors,
        variant,
        contrast_level,
        is_dark,
        harmonize_customs,
    ))
}
