use std::collections::HashMap;

use material_design_2::MaterialDesign2Scheme;
use material_design_3::MaterialDesign3Scheme;
use material_design_3_dynamic::{build_baseline, build_dynamic_scheme, build_swatches, Variant};
use q_style::{QScheme, SchemeSetting};
use swatch_style::{SwatchEntry, SwatchSchemeSetting};
use wasm_bindgen::{prelude::wasm_bindgen, JsValue};

use crate::errors;

pub mod material_design_2;
pub mod material_design_3;
pub mod material_design_3_dynamic;
pub mod q_style;
pub mod swatch_style;

pub trait SchemeExport {
    fn output_css_variables(&self) -> String;
    fn output_css_auto_scheme_variables(&self) -> String;
    fn output_scss_variables(&self) -> String;
    fn output_javascript_object(&self) -> String;
}

#[wasm_bindgen]
pub fn generate_material_design_3_scheme(
    source_color: &str,
    error_color: &str,
    custom_colors: JsValue,
) -> Result<JsValue, errors::ColorError> {
    let custom_colors: HashMap<String, String> = serde_wasm_bindgen::from_value(custom_colors)
        .map_err(|_| errors::ColorError::UnableToParseArgument)?;
    let mut scheme = MaterialDesign3Scheme::new(source_color, error_color)?;
    for (name, color) in custom_colors {
        scheme.add_custom_color(name, color)?;
    }
    Ok(serde_wasm_bindgen::to_value(&(
        scheme.clone(),
        scheme.output_css_variables(),
        scheme.output_css_auto_scheme_variables(),
        scheme.output_scss_variables(),
        scheme.output_javascript_object(),
    ))
    .map_err(|_| errors::ColorError::UnableToAssembleOutput)?)
}

#[wasm_bindgen]
pub fn generate_material_design_2_scheme(
    primary_color: &str,
    secondary_color: &str,
    error_color: &str,
    custom_colors: JsValue,
) -> Result<JsValue, errors::ColorError> {
    let custom_colors: HashMap<String, String> = serde_wasm_bindgen::from_value(custom_colors)
        .map_err(|_| errors::ColorError::UnableToParseArgument)?;
    let mut scheme = MaterialDesign2Scheme::new(primary_color, secondary_color, error_color)?;
    for (name, color) in custom_colors {
        scheme.add_custom_color(&name, &color)?;
    }
    Ok(serde_wasm_bindgen::to_value(&(
        scheme.clone(),
        scheme.output_css_variables(),
        scheme.output_css_auto_scheme_variables(),
        scheme.output_scss_variables(),
        scheme.output_javascript_object(),
    ))
    .map_err(|_| errors::ColorError::UnableToAssembleOutput)?)
}

#[wasm_bindgen]
pub fn generate_q_scheme_automatically(
    primary_color: &str,
    danger_color: &str,
    success_color: &str,
    warning_color: &str,
    info_color: &str,
    fg_color: &str,
    bg_color: &str,
    setting: SchemeSetting,
) -> Result<JsValue, errors::ColorError> {
    let scheme = QScheme::new(
        primary_color,
        danger_color,
        success_color,
        warning_color,
        info_color,
        fg_color,
        bg_color,
        setting,
    )?;
    Ok(serde_wasm_bindgen::to_value(&(
        scheme.clone(),
        scheme.output_css_variables(),
        scheme.output_css_auto_scheme_variables(),
        scheme.output_scss_variables(),
        scheme.output_javascript_object(),
    ))
    .map_err(|_| errors::ColorError::UnableToAssembleOutput)?)
}

#[wasm_bindgen]
pub fn generate_q_scheme_manually(
    primary_color: &str,
    secondary_color: Option<String>,
    tertiary_color: Option<String>,
    accent_color: Option<String>,
    danger_color: &str,
    success_color: &str,
    warning_color: &str,
    info_color: &str,
    fg_color: &str,
    bg_color: &str,
    setting: SchemeSetting,
) -> Result<JsValue, errors::ColorError> {
    let scheme = QScheme::custom(
        primary_color,
        secondary_color.as_deref(),
        tertiary_color.as_deref(),
        accent_color.as_deref(),
        danger_color,
        success_color,
        warning_color,
        info_color,
        fg_color,
        bg_color,
        setting,
    )?;
    Ok(serde_wasm_bindgen::to_value(&(
        scheme.clone(),
        scheme.output_css_variables(),
        scheme.output_css_auto_scheme_variables(),
        scheme.output_scss_variables(),
        scheme.output_javascript_object(),
    ))
    .map_err(|_| errors::ColorError::UnableToAssembleOutput)?)
}

#[wasm_bindgen]
pub fn generate_swatch_scheme(
    colors: Vec<SwatchEntry>,
    setting: SwatchSchemeSetting,
) -> Result<JsValue, errors::ColorError> {
    let scheme = swatch_style::SwatchScheme::new(colors, setting)?;
    Ok(serde_wasm_bindgen::to_value(&(
        scheme.swatches(),
        scheme.output_css_variables(),
        scheme.output_css_auto_scheme_variables(),
        scheme.output_scss_variables(),
        scheme.output_javascript_object(),
    ))
    .map_err(|_| errors::ColorError::UnableToAssembleOutput)?)
}

#[wasm_bindgen]
pub fn generate_material_design_3_dynamic_scheme(
    source_color: &str,
    error_color: Option<String>,
    variant: u8,
    contrast_level: f32,
    harmonize_customs: bool,
    custom_colors: JsValue,
) -> Result<JsValue, errors::ColorError> {
    let custom_colors: HashMap<String, String> = serde_wasm_bindgen::from_value(custom_colors)
        .map_err(|_| errors::ColorError::UnableToParseArgument)?;
    let variant = Variant::from_u8(variant);

    let light_scheme = build_dynamic_scheme(
        source_color,
        error_color.clone(),
        custom_colors.clone(),
        variant,
        contrast_level,
        false,
        harmonize_customs,
    )?;
    let dark_scheme = build_dynamic_scheme(
        source_color,
        error_color,
        custom_colors,
        variant,
        contrast_level,
        true,
        harmonize_customs,
    )?;

    let scheme = MaterialDesign3Scheme::full_custom(
        build_baseline(&light_scheme),
        build_baseline(&dark_scheme),
        build_swatches(&light_scheme),
    );

    Ok(serde_wasm_bindgen::to_value(&(
        scheme.clone(),
        scheme.output_css_variables(),
        scheme.output_css_auto_scheme_variables(),
        scheme.output_scss_variables(),
        scheme.output_javascript_object(),
    ))
    .map_err(|_| errors::ColorError::UnableToAssembleOutput)?)
}
