use linked_hash_map::LinkedHashMap;
use palette::{
    color_theory::{Analogous, Complementary, SplitComplementary, Tetradic, Triadic},
    Oklch, ShiftHue,
};
use serde::{ser::SerializeStruct, Serialize};

use crate::convert::map_oklch_to_srgb_hex;

use super::{
    color_set::ColorSet,
    neutral_swatch::NeutralSwatch,
    scheme_setting::{ColorExpand, SchemeSetting},
};

#[derive(Debug, Clone)]
pub struct Baseline {
    pub primary: ColorSet,
    pub secondary: Option<ColorSet>,
    pub tertiary: Option<ColorSet>,
    pub accent: Option<ColorSet>,
    pub neutral: ColorSet,
    pub danger: ColorSet,
    pub success: ColorSet,
    pub warning: ColorSet,
    pub info: ColorSet,
    pub outline: ColorSet,
    pub foreground: Oklch,
    pub background: Oklch,
    _neutral_swatch: NeutralSwatch,
}

impl Baseline {
    pub fn custom(
        primary: &Oklch,
        secondary: &Option<Oklch>,
        tertiary: &Option<Oklch>,
        accent: &Option<Oklch>,
        danger: &Oklch,
        success: &Oklch,
        warning: &Oklch,
        info: &Oklch,
        foreground: &Oklch,
        background: &Oklch,
        setting: SchemeSetting,
    ) -> Self {
        let neutral_swatch = NeutralSwatch::new(*foreground, *background);
        let neutral_color = neutral_swatch.get(primary.l);
        let outline_color = neutral_swatch.get(background.l * 0.7);

        Self {
            primary: ColorSet::new(primary, &neutral_swatch, foreground.l, &setting),
            secondary: secondary
                .map(|color| ColorSet::new(&color, &neutral_swatch, foreground.l, &setting)),
            tertiary: tertiary
                .map(|color| ColorSet::new(&color, &neutral_swatch, foreground.l, &setting)),
            accent: accent
                .map(|color| ColorSet::new(&color, &neutral_swatch, foreground.l, &setting)),
            neutral: ColorSet::new(&neutral_color, &neutral_swatch, foreground.l, &setting),
            danger: ColorSet::new(danger, &neutral_swatch, foreground.l, &setting),
            success: ColorSet::new(success, &neutral_swatch, foreground.l, &setting),
            warning: ColorSet::new(warning, &neutral_swatch, foreground.l, &setting),
            info: ColorSet::new(info, &neutral_swatch, foreground.l, &setting),
            outline: ColorSet::new(&outline_color, &neutral_swatch, foreground.l, &setting),
            foreground: *foreground,
            background: *background,
            _neutral_swatch: neutral_swatch,
        }
    }

    pub fn new(
        primary: &Oklch,
        danger: &Oklch,
        success: &Oklch,
        warning: &Oklch,
        info: &Oklch,
        foreground: &Oklch,
        background: &Oklch,
        setting: SchemeSetting,
    ) -> Self {
        let (secondary, tertiary, accent) = match setting.expand_method {
            ColorExpand::Complementary => (Some(primary.complementary()), None, None),
            ColorExpand::Analogous => {
                let analogous_color = primary.analogous();
                (Some(analogous_color.0), Some(analogous_color.1), None)
            }
            ColorExpand::AnalogousAndComplementary => {
                let analogous_color = primary.analogous();
                (
                    Some(analogous_color.0),
                    Some(analogous_color.1),
                    Some(primary.complementary()),
                )
            }
            ColorExpand::Triadic => {
                let triadic_color = primary.triadic();
                (Some(triadic_color.0), Some(triadic_color.1), None)
            }
            ColorExpand::SplitComplementary => {
                let split_complementary_color = primary.split_complementary();
                (
                    Some(split_complementary_color.0),
                    None,
                    Some(split_complementary_color.1),
                )
            }
            ColorExpand::Tetradic => {
                let tetradic_color = primary.tetradic();
                (
                    Some(tetradic_color.0),
                    Some(tetradic_color.2),
                    Some(tetradic_color.1),
                )
            }
            ColorExpand::Square => {
                let c_90 = primary.shift_hue(90.0);
                let complementary = primary.complementary();
                let c_270 = primary.shift_hue(270.0);
                (Some(c_90), Some(c_270), Some(complementary))
            }
        };

        Self::custom(
            primary, &secondary, &tertiary, &accent, danger, success, warning, info, foreground,
            background, setting,
        )
    }

    pub fn to_css_variables(&self, prefix: &str) -> Vec<String> {
        let mut variables = Vec::new();

        variables.extend(self.primary.to_css_variables(prefix, "primary"));
        if let Some(secondary) = &self.secondary {
            variables.extend(secondary.to_css_variables(prefix, "secondary"));
        }
        if let Some(tertiary) = &self.tertiary {
            variables.extend(tertiary.to_css_variables(prefix, "tertiary"));
        }
        if let Some(accent) = &self.accent {
            variables.extend(accent.to_css_variables(prefix, "accent"));
        }
        variables.extend(self.neutral.to_css_variables(prefix, "neutral"));
        variables.extend(self.danger.to_css_variables(prefix, "danger"));
        variables.extend(self.success.to_css_variables(prefix, "success"));
        variables.extend(self.warning.to_css_variables(prefix, "warning"));
        variables.extend(self.info.to_css_variables(prefix, "info"));
        variables.extend(self.outline.to_css_variables(prefix, "outline"));
        variables.push(format!(
            "--color-{}-foreground: #{};",
            prefix,
            map_oklch_to_srgb_hex(&self.foreground)
        ));
        variables.push(format!(
            "--color-{}-background: #{};",
            prefix,
            map_oklch_to_srgb_hex(&self.background)
        ));

        variables
    }

    pub fn to_css_auto_scheme_collection(&self) -> LinkedHashMap<String, String> {
        let mut collection = LinkedHashMap::new();

        collection.extend(self.primary.to_css_auto_scheme_collection("primary"));
        if let Some(secondary) = &self.secondary {
            collection.extend(secondary.to_css_auto_scheme_collection("secondary"));
        }
        if let Some(tertiary) = &self.tertiary {
            collection.extend(tertiary.to_css_auto_scheme_collection("tertiary"));
        }
        if let Some(accent) = &self.accent {
            collection.extend(accent.to_css_auto_scheme_collection("accent"));
        }
        collection.extend(self.neutral.to_css_auto_scheme_collection("neutral"));
        collection.extend(self.danger.to_css_auto_scheme_collection("danger"));
        collection.extend(self.success.to_css_auto_scheme_collection("success"));
        collection.extend(self.warning.to_css_auto_scheme_collection("warning"));
        collection.extend(self.info.to_css_auto_scheme_collection("info"));
        collection.extend(self.outline.to_css_auto_scheme_collection("outline"));
        collection.insert(
            "foreground".to_string(),
            map_oklch_to_srgb_hex(&self.foreground),
        );
        collection.insert(
            "background".to_string(),
            map_oklch_to_srgb_hex(&self.background),
        );

        collection
    }

    pub fn to_scss_variables(&self, prefix: &str) -> Vec<String> {
        let mut variables = Vec::new();

        variables.extend(self.primary.to_scss_variables(prefix, "primary"));
        if let Some(secondary) = &self.secondary {
            variables.extend(secondary.to_scss_variables(prefix, "secondary"));
        }
        if let Some(tertiary) = &self.tertiary {
            variables.extend(tertiary.to_scss_variables(prefix, "tertiary"));
        }
        if let Some(accent) = &self.accent {
            variables.extend(accent.to_scss_variables(prefix, "accent"));
        }
        variables.extend(self.neutral.to_scss_variables(prefix, "neutral"));
        variables.extend(self.danger.to_scss_variables(prefix, "danger"));
        variables.extend(self.success.to_scss_variables(prefix, "success"));
        variables.extend(self.warning.to_scss_variables(prefix, "warning"));
        variables.extend(self.info.to_scss_variables(prefix, "info"));
        variables.extend(self.outline.to_scss_variables(prefix, "outline"));
        variables.push(format!(
            "$color-{}-foreground: #{};",
            prefix,
            map_oklch_to_srgb_hex(&self.foreground)
        ));
        variables.push(format!(
            "$color-{}-background: #{};",
            prefix,
            map_oklch_to_srgb_hex(&self.background)
        ));

        variables
    }

    pub fn to_javascript_fields(&self) -> Vec<String> {
        let mut variables = Vec::new();

        variables.extend(self.primary.to_javascript_fields("primary"));
        if let Some(secondary) = &self.secondary {
            variables.extend(secondary.to_javascript_fields("secondary"));
        }
        if let Some(tertiary) = &self.tertiary {
            variables.extend(tertiary.to_javascript_fields("tertiary"));
        }
        if let Some(accent) = &self.accent {
            variables.extend(accent.to_javascript_fields("accent"));
        }
        variables.extend(self.neutral.to_javascript_fields("neutral"));
        variables.extend(self.danger.to_javascript_fields("danger"));
        variables.extend(self.success.to_javascript_fields("success"));
        variables.extend(self.warning.to_javascript_fields("warning"));
        variables.extend(self.info.to_javascript_fields("info"));
        variables.extend(self.outline.to_javascript_fields("outline"));
        variables.push(format!(
            "foreground: '#{}',",
            map_oklch_to_srgb_hex(&self.foreground)
        ));
        variables.push(format!(
            "background: '#{}',",
            map_oklch_to_srgb_hex(&self.background)
        ));

        variables
    }
}

impl Serialize for Baseline {
    fn serialize<S: serde::Serializer>(&self, serializer: S) -> Result<S::Ok, S::Error> {
        let mut map = serializer.serialize_struct("Baseline", 13)?;

        map.serialize_field("primary", &self.primary)?;
        map.serialize_field("secondary", &self.secondary)?;
        map.serialize_field("tertiary", &self.tertiary)?;
        map.serialize_field("accent", &self.accent)?;
        map.serialize_field("neutral", &self.neutral)?;
        map.serialize_field("danger", &self.danger)?;
        map.serialize_field("success", &self.success)?;
        map.serialize_field("warning", &self.warning)?;
        map.serialize_field("info", &self.info)?;
        map.serialize_field("outline", &self.outline)?;
        map.serialize_field("foreground", &map_oklch_to_srgb_hex(&self.foreground))?;
        map.serialize_field("background", &map_oklch_to_srgb_hex(&self.background))?;
        map.end()
    }
}
