use linked_hash_map::LinkedHashMap;
use palette::{color_difference::Wcag21RelativeContrast, luma::Luma, Oklch};
use serde::{ser::SerializeStruct, Serialize};

use crate::convert::{map_oklch_to_luma, map_oklch_to_srgb_hex};

use super::{
    neutral_swatch::NeutralSwatch,
    scheme_setting::{SchemeSetting, WACGSetting},
};

#[derive(Debug, Clone)]
pub struct ColorSet {
    pub root: Oklch,
    pub hover: Oklch,
    pub active: Oklch,
    pub focus: Oklch,
    pub disabled: Oklch,
    pub on_root: Oklch,
    pub on_hover: Oklch,
    pub on_active: Oklch,
    pub on_focus: Oklch,
    pub on_disabled: Oklch,
}

fn fit_to_wacg(reference: &Oklch, neutral_swatch: &NeutralSwatch, ratio: f32) -> Oklch {
    let reference_luma = map_oklch_to_luma(reference);

    let match_wacg = |original: &Oklch<f32>, reference: &Luma| {
        let luma = map_oklch_to_luma(original);
        luma.relative_contrast(*reference)
    };

    let mut fit_contrast = (f32::INFINITY, f32::NEG_INFINITY);
    let mut closest_contrast = (f32::INFINITY, f32::NEG_INFINITY);
    for scan_lightness in (0..=100).map(|x| x as f32 / 100.0) {
        let new_target = neutral_swatch.get(scan_lightness);
        let contrast_ratio = match_wacg(&new_target, &reference_luma);
        if (contrast_ratio - ratio).abs() < (closest_contrast.0 - ratio).abs()
            && scan_lightness > closest_contrast.1
        {
            closest_contrast = (contrast_ratio, scan_lightness);
        }
        if contrast_ratio >= ratio
            && (contrast_ratio - ratio).abs() < (closest_contrast.0 - ratio).abs()
        {
            fit_contrast = (contrast_ratio, scan_lightness);
        }
    }

    neutral_swatch.get(if fit_contrast.0 == f32::INFINITY {
        closest_contrast.1
    } else {
        fit_contrast.1
    })
}

impl ColorSet {
    pub fn new(
        color: &Oklch,
        neutral_swatch: &NeutralSwatch,
        foreground_lightness: f32,
        setting: &SchemeSetting,
    ) -> Self {
        let root = color.clone();
        let hover = color * setting.hover;
        let active = color * setting.active;
        let focus = color * setting.focus;
        let disabled = color * setting.disabled;

        let (on_root, on_hover, on_active, on_focus, on_disabled) = match setting.wacg_follows {
            WACGSetting::Fixed => (
                neutral_swatch.get(foreground_lightness),
                neutral_swatch.get(foreground_lightness),
                neutral_swatch.get(foreground_lightness),
                neutral_swatch.get(foreground_lightness),
                neutral_swatch.get(foreground_lightness),
            ),
            WACGSetting::AutomaticAA => (
                fit_to_wacg(&root, neutral_swatch, 4.5),
                fit_to_wacg(&hover, neutral_swatch, 4.5),
                fit_to_wacg(&active, neutral_swatch, 4.5),
                fit_to_wacg(&focus, neutral_swatch, 4.5),
                fit_to_wacg(&disabled, neutral_swatch, 4.5),
            ),
            WACGSetting::AutomaticAAA => (
                fit_to_wacg(&root, neutral_swatch, 7.0),
                fit_to_wacg(&hover, neutral_swatch, 7.0),
                fit_to_wacg(&active, neutral_swatch, 7.0),
                fit_to_wacg(&focus, neutral_swatch, 7.0),
                fit_to_wacg(&disabled, neutral_swatch, 7.0),
            ),
            WACGSetting::HighContrast => (
                fit_to_wacg(&root, neutral_swatch, 21.0),
                fit_to_wacg(&hover, neutral_swatch, 21.0),
                fit_to_wacg(&active, neutral_swatch, 21.0),
                fit_to_wacg(&focus, neutral_swatch, 21.0),
                fit_to_wacg(&disabled, neutral_swatch, 21.0),
            ),
        };

        Self {
            root,
            hover,
            active,
            focus,
            disabled,
            on_root,
            on_hover,
            on_active,
            on_focus,
            on_disabled,
        }
    }

    pub fn to_css_variables(&self, prefix: &str, name: &str) -> Vec<String> {
        let mut variables = Vec::new();

        variables.push(format!(
            "--color-{}-{}: #{};",
            prefix,
            name,
            map_oklch_to_srgb_hex(&self.root)
        ));
        variables.push(format!(
            "--color-{}-{}-hover: #{};",
            prefix,
            name,
            map_oklch_to_srgb_hex(&self.hover)
        ));
        variables.push(format!(
            "--color-{}-{}-active: #{};",
            prefix,
            name,
            map_oklch_to_srgb_hex(&self.active)
        ));
        variables.push(format!(
            "--color-{}-{}-focus: #{};",
            prefix,
            name,
            map_oklch_to_srgb_hex(&self.focus)
        ));
        variables.push(format!(
            "--color-{}-{}-disabled: #{};",
            prefix,
            name,
            map_oklch_to_srgb_hex(&self.disabled)
        ));
        variables.push(format!(
            "--color-{}-on-{}: #{};",
            prefix,
            name,
            map_oklch_to_srgb_hex(&self.on_root)
        ));
        variables.push(format!(
            "--color-{}-on-{}-hover: #{};",
            prefix,
            name,
            map_oklch_to_srgb_hex(&self.on_hover)
        ));
        variables.push(format!(
            "--color-{}-on-{}-active: #{};",
            prefix,
            name,
            map_oklch_to_srgb_hex(&self.on_active)
        ));
        variables.push(format!(
            "--color-{}-on-{}-focus: #{};",
            prefix,
            name,
            map_oklch_to_srgb_hex(&self.on_focus)
        ));
        variables.push(format!(
            "--color-{}-on-{}-disabled: #{};",
            prefix,
            name,
            map_oklch_to_srgb_hex(&self.on_disabled)
        ));

        variables
    }

    pub fn to_css_auto_scheme_collection(&self, name: &str) -> LinkedHashMap<String, String> {
        let mut collection = LinkedHashMap::new();

        collection.insert(format!("{}", name), map_oklch_to_srgb_hex(&self.root));
        collection.insert(
            format!("{}-hover", name),
            map_oklch_to_srgb_hex(&self.hover),
        );
        collection.insert(
            format!("{}-active", name),
            map_oklch_to_srgb_hex(&self.active),
        );
        collection.insert(
            format!("{}-focus", name),
            map_oklch_to_srgb_hex(&self.focus),
        );
        collection.insert(
            format!("{}-disabled", name),
            map_oklch_to_srgb_hex(&self.disabled),
        );
        collection.insert(format!("on-{}", name), map_oklch_to_srgb_hex(&self.on_root));
        collection.insert(
            format!("on-{}-hover", name),
            map_oklch_to_srgb_hex(&self.on_hover),
        );
        collection.insert(
            format!("on-{}-active", name),
            map_oklch_to_srgb_hex(&self.on_active),
        );
        collection.insert(
            format!("on-{}-focus", name),
            map_oklch_to_srgb_hex(&self.on_focus),
        );
        collection.insert(
            format!("on-{}-disabled", name),
            map_oklch_to_srgb_hex(&self.on_disabled),
        );

        collection
    }

    pub fn to_scss_variables(&self, prefix: &str, name: &str) -> Vec<String> {
        let mut variables = Vec::new();

        variables.push(format!(
            "$color-{}-{}: #{};",
            prefix,
            name,
            map_oklch_to_srgb_hex(&self.root)
        ));
        variables.push(format!(
            "$color-{}-{}-hover: #{};",
            prefix,
            name,
            map_oklch_to_srgb_hex(&self.hover)
        ));
        variables.push(format!(
            "$color-{}-{}-active: #{};",
            prefix,
            name,
            map_oklch_to_srgb_hex(&self.active)
        ));
        variables.push(format!(
            "$color-{}-{}-focus: #{};",
            prefix,
            name,
            map_oklch_to_srgb_hex(&self.focus)
        ));
        variables.push(format!(
            "$color-{}-{}-disabled: #{};",
            prefix,
            name,
            map_oklch_to_srgb_hex(&self.disabled)
        ));
        variables.push(format!(
            "$color-{}-on-{}: #{};",
            prefix,
            name,
            map_oklch_to_srgb_hex(&self.on_root)
        ));
        variables.push(format!(
            "$color-{}-on-{}-hover: #{};",
            prefix,
            name,
            map_oklch_to_srgb_hex(&self.on_hover)
        ));
        variables.push(format!(
            "$color-{}-on-{}-active: #{};",
            prefix,
            name,
            map_oklch_to_srgb_hex(&self.on_active)
        ));
        variables.push(format!(
            "$color-{}-on-{}-focus: #{};",
            prefix,
            name,
            map_oklch_to_srgb_hex(&self.on_focus)
        ));
        variables.push(format!(
            "$color-{}-on-{}-disabled: #{};",
            prefix,
            name,
            map_oklch_to_srgb_hex(&self.on_disabled)
        ));

        variables
    }

    pub fn to_javascript_fields(&self, name: &str) -> Vec<String> {
        let mut variables = Vec::new();
        let capitalized_name = name
            .chars()
            .next()
            .unwrap()
            .to_ascii_uppercase()
            .to_string()
            + &name[1..];

        variables.push(format!(
            "{}: '#{}',",
            name,
            map_oklch_to_srgb_hex(&self.root)
        ));
        variables.push(format!(
            "{}Hover: '#{}',",
            name,
            map_oklch_to_srgb_hex(&self.hover)
        ));
        variables.push(format!(
            "{}Active: '#{}',",
            name,
            map_oklch_to_srgb_hex(&self.active)
        ));
        variables.push(format!(
            "{}Focus: '#{}',",
            name,
            map_oklch_to_srgb_hex(&self.focus)
        ));
        variables.push(format!(
            "{}Disabled: '#{}',",
            name,
            map_oklch_to_srgb_hex(&self.disabled)
        ));
        variables.push(format!(
            "on{}: '#{}',",
            capitalized_name,
            map_oklch_to_srgb_hex(&self.on_root)
        ));
        variables.push(format!(
            "on{}Hover: '#{}',",
            capitalized_name,
            map_oklch_to_srgb_hex(&self.on_hover)
        ));
        variables.push(format!(
            "on{}Active: '#{}',",
            capitalized_name,
            map_oklch_to_srgb_hex(&self.on_active)
        ));
        variables.push(format!(
            "on{}Focus: '#{}',",
            capitalized_name,
            map_oklch_to_srgb_hex(&self.on_focus)
        ));
        variables.push(format!(
            "on{}Disabled: '#{}',",
            capitalized_name,
            map_oklch_to_srgb_hex(&self.on_disabled)
        ));

        variables
    }
}

impl Serialize for ColorSet {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: serde::Serializer,
    {
        let root = map_oklch_to_srgb_hex(&self.root);
        let hover = map_oklch_to_srgb_hex(&self.hover);
        let active = map_oklch_to_srgb_hex(&self.active);
        let focus = map_oklch_to_srgb_hex(&self.focus);
        let disabled = map_oklch_to_srgb_hex(&self.disabled);
        let on_root = map_oklch_to_srgb_hex(&self.on_root);
        let on_hover = map_oklch_to_srgb_hex(&self.on_hover);
        let on_active = map_oklch_to_srgb_hex(&self.on_active);
        let on_focus = map_oklch_to_srgb_hex(&self.on_focus);
        let on_disabled = map_oklch_to_srgb_hex(&self.on_disabled);

        let mut state = serializer.serialize_struct("ColorSet", 10)?;
        state.serialize_field("root", &root)?;
        state.serialize_field("hover", &hover)?;
        state.serialize_field("active", &active)?;
        state.serialize_field("focus", &focus)?;
        state.serialize_field("disabled", &disabled)?;
        state.serialize_field("onRoot", &on_root)?;
        state.serialize_field("onHover", &on_hover)?;
        state.serialize_field("onActive", &on_active)?;
        state.serialize_field("onFocus", &on_focus)?;
        state.serialize_field("onDisabled", &on_disabled)?;
        state.end()
    }
}
