use linked_hash_map::LinkedHashMap;
use palette::Oklch;

use crate::convert::map_oklch_to_srgb_hex;

use super::setting::SwatchSchemeSetting;

#[derive(Debug, Clone)]
pub struct Swatch {
    min_key: f32,
    max_key: f32,
    primary_key: Oklch,
    include_primary: bool,
    color_amount: usize,
}

impl Swatch {
    pub fn new(primary: &Oklch, setting: &SwatchSchemeSetting) -> Self {
        Self {
            min_key: primary.l.min(setting.min_lightness),
            max_key: primary.l.max(setting.max_lightness),
            primary_key: primary.clone(),
            include_primary: setting.include_primary,
            color_amount: setting.amount,
        }
    }

    fn find_interval(&self) -> usize {
        if !self.include_primary {
            return 0;
        }
        if self.primary_key.l == self.min_key {
            return 0;
        }
        if self.primary_key.l == self.max_key {
            return self.color_amount - 1;
        }
        let step = (self.max_key - self.min_key) / (self.color_amount - 1) as f32;
        ((self.primary_key.l - self.min_key) / step).ceil() as usize
    }

    pub fn swatch(&self) -> Vec<Oklch> {
        let mut swatch = Vec::new();
        if self.include_primary {
            let primary_index = self.find_interval();
            if primary_index > 0 {
                let step = (self.primary_key.l - self.min_key) / (primary_index - 1) as f32;
                for i in 0..primary_index {
                    let lightness = self.min_key + step * i as f32;
                    swatch.push(Oklch {
                        l: lightness,
                        ..self.primary_key
                    });
                }
            }
            if primary_index < self.color_amount - 1 {
                let step = (self.max_key - self.primary_key.l)
                    / (self.color_amount - primary_index - 1) as f32;
                for i in primary_index..self.color_amount {
                    let lightness = self.min_key + step * i as f32;
                    swatch.push(Oklch {
                        l: lightness,
                        ..self.primary_key
                    });
                }
            }
        } else {
            let step = (self.max_key - self.min_key) / (self.color_amount - 1) as f32;
            for i in 0..self.color_amount {
                let lightness = self.min_key + step * i as f32;
                swatch.push(Oklch {
                    l: lightness,
                    ..self.primary_key
                });
            }
        }
        swatch
    }

    pub fn swtch_hex(&self) -> Vec<String> {
        self.swatch().iter().map(map_oklch_to_srgb_hex).collect()
    }

    pub fn to_css_variables(&self, prefix: &str, name: &str) -> Vec<String> {
        let mut variables = Vec::new();
        for (i, color) in self.swatch().iter().enumerate() {
            variables.push(format!(
                "--color-{}-{}-{}: #{};",
                prefix,
                name,
                i * 100,
                map_oklch_to_srgb_hex(color)
            ));
        }
        variables
    }

    pub fn to_css_auto_scheme_collection(&self, name: &str) -> LinkedHashMap<String, String> {
        let mut collection = LinkedHashMap::new();
        for (i, color) in self.swatch().iter().enumerate() {
            collection.insert(
                format!("{}-{}", name, i * 100),
                map_oklch_to_srgb_hex(color),
            );
        }
        collection
    }

    pub fn to_scss_variables(&self, prefix: &str, name: &str) -> Vec<String> {
        let mut variables = Vec::new();
        for (i, color) in self.swatch().iter().enumerate() {
            variables.push(format!(
                "${}-{}-{}: #{};",
                prefix,
                name,
                i * 100,
                map_oklch_to_srgb_hex(color)
            ));
        }
        variables
    }

    pub fn to_javascript_fields(&self, prefix: &str, name: &str) -> Vec<String> {
        let mut variables = Vec::new();
        let capitalized_name = name
            .chars()
            .next()
            .unwrap()
            .to_ascii_uppercase()
            .to_string()
            + &name[1..];
        for (i, color) in self.swatch().iter().enumerate() {
            variables.push(format!(
                "{}{}{}: '#{}',",
                prefix,
                capitalized_name,
                i * 100,
                map_oklch_to_srgb_hex(color)
            ));
        }
        variables
    }
}
