import { includes, isEmpty, isNil } from 'lodash-es';
import { useActionState, useCallback, useMemo, useState } from 'react';
import { useColorFunction } from '../../../ColorFunctionContext';
import { FloatColorPicker } from '../../../components/FloatColorPicker';
import { NotificationType, useNotification } from '../../../components/Notifications';
import { ScrollArea } from '../../../components/ScrollArea';
import {
  MaterialDesign3Scheme,
  MaterialDesign3SchemeSource,
  MaterialDesign3SchemeStorage,
} from '../../../material-3-scheme';
import { SchemeContent } from '../../../models';
import { useUpdateScheme } from '../../../stores/schemes';
import { mapToObject } from '../../../utls';
import { ColorEntry, IdenticalColorEntry } from '../ColorEntry';
import styles from './Builder.module.css';

type M3SchemeBuilderProps = {
  scheme: SchemeContent<MaterialDesign3SchemeStorage>;
  onBuildCompleted?: () => void;
};

export function M3SchemeBuilder({ scheme, onBuildCompleted }: M3SchemeBuilderProps) {
  const { showToast } = useNotification();
  const { colorFn } = useColorFunction();
  const updateScheme = useUpdateScheme(scheme.id);
  const originalColors = useMemo(() => {
    return Object.entries(scheme.schemeStorage.source?.custom_colors ?? {}).map(
      ([name, color], index) => ({ id: `oc_${index}`, name, color } as IdenticalColorEntry),
    );
  }, [scheme.schemeStorage.source]);
  const [newColors, setNewColors] = useState<IdenticalColorEntry[]>([]);
  const [deleted, setDeleted] = useState<string[]>([]);
  const addEntryAction = useCallback(() => {
    setNewColors((prev) => [...prev, { id: `nc_${prev.length}`, name: '', color: '' }]);
  }, []);
  const colorKeys = useMemo(
    () =>
      [...originalColors, ...newColors]
        .map((color) => color.id)
        .filter((c) => !includes(deleted, c)),
    [originalColors, newColors, deleted],
  );

  const collectSchemeSource = (formData: FormData): MaterialDesign3SchemeSource => {
    const sourceColor = formData.get('source') as string;
    const errorColor = formData.get('error') as string;
    const customColors: Record<string, string> = {};
    for (const key of colorKeys) {
      const name = formData.get(`name_${key}`) as string;
      const color = formData.get(`color_${key}`) as string;
      if (isNil(name) || isEmpty(name) || isNil(color) || isEmpty(color)) continue;
      customColors[name] = color;
    }

    return {
      source: isNil(sourceColor) || isEmpty(sourceColor) ? null : sourceColor,
      error: isNil(errorColor) || isEmpty(errorColor) ? null : errorColor,
      custom_colors: customColors,
    };
  };

  const [, handleDraftAction] = useActionState<Map<string, string>, FormData>(
    (_state, formData) => {
      const errMsg = new Map<string, string>();

      const collectedSource = collectSchemeSource(formData);
      updateScheme((prev) => {
        prev.schemeStorage.source = collectedSource;
        return prev;
      });
      setNewColors([]);

      showToast(NotificationType.SUCCESS, 'Scheme draft saved!', 'tabler:device-floppy', 3000);

      return errMsg;
    },
    new Map<string, string>(),
  );
  const [errMsg, handleSubmitAction] = useActionState<Map<string, string>, FormData>(
    (_state, formData) => {
      const errMsg = new Map<string, string>();

      try {
        const collectedSource = collectSchemeSource(formData);
        if (isNil(collectedSource.source) || isEmpty(collectedSource.source)) {
          errMsg.set('source', 'Source color is required');
        }
        if (isNil(collectedSource.error) || isEmpty(collectedSource.error)) {
          errMsg.set('error', 'Error color is required');
        }
        if (!isEmpty(errMsg)) return errMsg;

        const generatedScheme = colorFn?.generate_material_design_3_scheme(
          collectedSource.source,
          collectedSource.error,
          collectedSource.custom_colors,
        );
        updateScheme((prev) => {
          prev.schemeStorage.source = collectedSource;
          prev.schemeStorage.scheme = {
            white: generatedScheme[0].white,
            black: generatedScheme[0].black,
            light_baseline: {
              ...generatedScheme[0].light_baseline,
              customs: mapToObject(generatedScheme[0].light_baseline.customs),
            },
            dark_baseline: {
              ...generatedScheme[0].dark_baseline,
              customs: mapToObject(generatedScheme[0].dark_baseline.customs),
            },
          } as MaterialDesign3Scheme;
          prev.schemeStorage.cssVariables = generatedScheme[1];
          prev.schemeStorage.cssAutoSchemeVariables = generatedScheme[2];
          prev.schemeStorage.scssVariables = generatedScheme[3];
          prev.schemeStorage.jsVariables = generatedScheme[4];
          return prev;
        });

        onBuildCompleted?.();
      } catch (e) {
        console.error('[generate m3 scheme]', e);
      }

      return errMsg;
    },
    new Map<string, string>(),
  );

  return (
    <ScrollArea enableY>
      <form action={handleSubmitAction} className={styles.builder_layout}>
        <h5 className={styles.segment_title}>Required Colors</h5>
        <label className={styles.label}>Source Color</label>
        <div className={styles.color_picker_row}>
          <FloatColorPicker
            name="source"
            color={
              isNil(scheme.schemeStorage.source?.source) ||
              isEmpty(scheme.schemeStorage.source?.source)
                ? undefined
                : scheme.schemeStorage.source?.source
            }
          />
          {errMsg.has('source') && <span className={styles.error_msg}>{errMsg.get('source')}</span>}
        </div>
        <label className={styles.label}>Error Color</label>
        <div className={styles.color_picker_row}>
          <FloatColorPicker
            name="error"
            color={
              isNil(scheme.schemeStorage.source?.error) ||
              isEmpty(scheme.schemeStorage.source.error)
                ? undefined
                : scheme.schemeStorage.source.error
            }
          />
          {errMsg.has('error') && <span className={styles.error_msg}>{errMsg.get('error')}</span>}
        </div>
        <h5 className={styles.segment_title}>Custom Colors</h5>
        <label style={{ gridColumn: 1 }}>Name</label>
        <label>Color</label>
        <div>
          <button type="button" className="small" onClick={addEntryAction}>
            Add Color
          </button>
        </div>
        {originalColors
          .filter((color) => !includes(deleted, color.id))
          .map((color) => (
            <ColorEntry
              key={color.id}
              entry={color}
              onDelete={(index) => setDeleted((prev) => [...prev, index])}
            />
          ))}
        {newColors
          .filter((color) => !includes(deleted, color.id))
          .map((color) => (
            <ColorEntry
              key={color.id}
              entry={color}
              onDelete={(index) => setDeleted((prev) => [...prev, index])}
            />
          ))}
        <div className={styles.button_row} style={{ gridColumn: '2 / span 2' }}>
          <button type="submit" className="primary">
            Build Scheme
          </button>
          <button type="submit" className="secondary" formAction={handleDraftAction}>
            Save Draft
          </button>
        </div>
      </form>
    </ScrollArea>
  );
}
